/*
 * Amanda, The Advanced Maryland Automatic Network Disk Archiver
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: James da Silva, Systems Design and Analysis Group
 *			   Computer Science Department
 *			   University of Maryland at College Park
 */
#include "amanda.h"
#include "conffile.h"
#include "infofile.h"
#include "logfile.h"
#include "stream.h"
#include "clock.h"

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#endif

#define CONNECT_TIMEOUT	5*60
#define READ_TIMEOUT	10*60
#define MAX_LINE 	1024
#define MAX_ARGS 	7
#define DATABUF_SIZE 	32*1024
#define MESGBUF_SIZE  	4*1024

#define SENDDUMP_PORT_DEFAULT	10070

#ifndef INSTRUMENTATION
#   define instr_start()
#   define instr_recvd(n)
#   define instr_end()
#endif


typedef enum { BOGUS, FILE_DUMP, PORT_DUMP, CONTINUE, ABORT, QUIT } cmd_t;

char *pname = "dumper";

char line[MAX_LINE];
char *argv[MAX_ARGS];
int argc;
int interactive;
char *handle;
int data_port, data_socket, client_socket;
char loginid[80];

char databuf[DATABUF_SIZE];
char mesgbuf[MESGBUF_SIZE];
char errstr[256];
char *dataptr;		/* data buffer markers */
int spaceleft, abort_pending, pid;
long dumpsize, origsize;
times_t runtime;

char errfname[MAX_LINE];
FILE *errf;
char *hostname, *diskname;
int level;
char datestamp[80];
time_t timestamp;

/* local functions */
static cmd_t getcmd P((void));
static void putresult();
static int run_dump P((char *host, int port, char *dsk, char *lev, char *opt));
static void do_dump P((int errfd, int infd, int outfd));

static void construct_datestamp(buf)
char *buf;
{
    struct tm *tm;

    timestamp = time((time_t *)NULL);
    tm = localtime(&timestamp);
    sprintf(buf, "%04d%02d%02d", tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday);
}

void main()
{
    cmd_t cmd;
    int infd, outfd, errfd, taper_port;
    struct passwd *pwptr;

    erroutput_type = (ERR_AMANDALOG|ERR_INTERACTIVE);

    if(read_conffile(CONFFILE_NAME)) 
	error("could not read conf file");

    /* setup server port first thing */

    if((data_socket = stream_server(&data_port)) == -1)
	error("could not setup server socket: %s", strerror(errno));

    if(geteuid() == 0) {
	/* set both real and effective uid's to real uid, likewise for gid */
	setgid(getgid());
	setuid(getuid());
    }
#ifdef BSD_SECURITY
    else error("must be run setuid root to communicate correctly");
#endif

    fprintf(stderr, "dumper: pid %d starting up, using port %d\n", 
	    getpid(), data_port);
    fflush(stderr);

    /* now, find out who I'm running as */

    if((pwptr = getpwuid(getuid())) == NULL)
	error("can't get login name for my uid %d", getuid());
    strcpy(loginid, pwptr->pw_name);

    interactive = isatty(0);
    pid = getpid();

    construct_datestamp(datestamp);

    do {
	cmd = getcmd();

	switch(cmd) {
	case QUIT:
	    break;
	case FILE_DUMP:

	    assert(argc == 7);
	    handle = argv[1];
	    hostname = argv[3];
	    diskname = argv[4];
	    level = atoi(argv[5]);

	    if((outfd = open(argv[2], O_WRONLY|O_CREAT, 0666)) == -1) {
		putresult("FAILED %s [file open: %s]\n",
			  handle, strerror(errno));
		break;
	    }

	    errfd = run_dump(hostname, data_port, diskname, argv[5], argv[6]);
	    if(errfd == -1) {
		close(outfd);
		break;
	    }

	    infd = stream_accept(data_socket, CONNECT_TIMEOUT,
				 DEFAULT_SIZE, DATABUF_SIZE);
	    if(infd == -1) {
		putresult("FATAL-TRY-AGAIN %s [accept timeout]\n", handle);
		error("pid %d bailing out on accept timeout, port %d", 
		      getpid(), data_port);
/*
		close(outfd);
		close(errfd);
		break;
*/
	    }

	    abort_pending = 0;
	    do_dump(errfd, infd, outfd);
	    close(errfd);
	    close(infd);
	    close(outfd);
	    if(abort_pending) putresult("ABORT-FINISHED %s\n", handle);
	    break;

	case PORT_DUMP:

	    /*
	     * Originally, we were planning to pass the taper port directly
	     * to the slave and just handle the dump stderr output here.
	     * This led to problems: we have no reliable way to time out on
	     * the stderr output.  Unix dump sends a status msg approx every
	     * five minutes, but other backup systems won't.  Also, the taper
	     * could time out, and we'd need some protocol to get that info
	     * to the dumper so that we don't hang.  Also, it's the dumper
	     * that updates the size statistics, so we'd need special case
	     * code in the taper for that case too.
	     *
	     * It's must simpler to just receive the data here and pass it
	     * to the taper ourselves.  It's an extra two copies, but that
	     * shouldn't be the bottleneck anyway.
	     */

	    assert(argc == 7);
	    handle = argv[1];
	    taper_port = atoi(argv[2]);
	    hostname = argv[3];
	    diskname = argv[4];
	    level = atoi(argv[5]);

	    /* connect outf to taper port */

	    outfd = stream_client("localhost", taper_port, 
				  DATABUF_SIZE, DEFAULT_SIZE);
	    if(outfd == -1) {
		putresult("FAILED %s [taper port open: %s]\n",
			  handle, strerror(errno));
		break;
	    }

	    errfd = run_dump(hostname, data_port, diskname, argv[5], argv[6]);
	    if(errfd == -1) break;

	    infd = stream_accept(data_socket, CONNECT_TIMEOUT,
				 DEFAULT_SIZE, DATABUF_SIZE);
	    if(infd == -1) {
		putresult("FATAL-TRY-AGAIN %s [accept timeout]\n", handle);
		error("pid %d bailing out on accept timeout, port %d", 
		      getpid(), data_port);
/*
		close(outfd);
		close(errfd);
		break;
*/
	    }

	    do_dump(errfd, infd, outfd);
	    close(errfd);
	    close(infd);
	    close(outfd);
	    break;

	default:
	    putresult("BAD-COMMAND %s\n", argv[0]);
	}
    } while(cmd != QUIT);
    exit(0);
}

static cmd_t getcmd()
{
    char *p;
    int arg;

    if(interactive) {
	printf("%s> ", pname); fflush(stdout);
    }

    if(fgets(line, MAX_LINE, stdin) == NULL)
	return QUIT;

    p = line;
    argc = 0;
    while(*p) {
	while(isspace(*p)) p++;
	if(argc < MAX_ARGS) argv[argc++] = p;
	while(*p && !isspace(*p)) p++;
	if(*p) *p++ = '\0';
    }
    for(arg = argc; arg < MAX_ARGS; arg++) argv[arg] = "";

#if DEBUG
    printf("argc = %d\n", argc);
    for(arg = 0; arg < MAX_ARGS; arg++)
	printf("argv[%d] = \"%s\"\n", arg, argv[arg]);
#endif

    /* not enough commands for a table lookup */

    if(!strcmp(argv[0],"FILE-DUMP")) return FILE_DUMP;
    else if(!strcmp(argv[0],"PORT-DUMP")) return PORT_DUMP;
    else if(!strcmp(argv[0],"CONTINUE")) return CONTINUE;
    else if(!strcmp(argv[0],"ABORT")) return ABORT;
    else if(!strcmp(argv[0],"QUIT")) return QUIT;
    return BOGUS;
}


static void putresult(format, va_alist)
char *format;
va_dcl
{
    va_list argp;
    char result[MAX_LINE];

    va_start(argp);
    vsprintf(result, format, argp);
    va_end(argp);
/*    fprintf(stderr, "dumper %d: putting result: %s", pid, result);
    fflush(stderr);*/
    write(1, result, strlen(result));
}


int update_dataptr(outf, size)
int outf, size;
/*
 * Updates the buffer pointer for the input data buffer.  The buffer is
 * written if it is full, or the remainder is zeroed if at eof.
 */
{
    int rc;
    cmd_t cmd;
    off_t pos;
    
    spaceleft -= size;
    dataptr += size;

    if(size == 0) {	/* eof, zero rest of buffer */
	memset(dataptr, '\0', spaceleft);
	spaceleft = 0;
    }

    if(spaceleft == 0) {	/* buffer is full, write it */
	pos = lseek(outf, 0L, SEEK_CUR);
	while((rc = write(outf, databuf, DATABUF_SIZE)) < DATABUF_SIZE) {
	    if(rc >= 0) {
		/*
		 * Assuming this means we ran out of space part way through,
		 * go back to start of block and try again.
		 *
		 * This assumption may be false if we are catching signals.
		 */
		lseek(outf, pos, SEEK_SET);

	    } else if(errno != ENOSPC) {
		putresult("FAILED %s [data write: %s]\n",
			  handle, strerror(errno));
		return 1;
	    }
	    putresult("NO-ROOM %s\n", handle);
	    cmd = getcmd();
	    assert(cmd == CONTINUE || cmd == ABORT);
	    if(cmd == CONTINUE) continue;
	    abort_pending = 1;
	    return 1;
	}
	spaceleft = DATABUF_SIZE;
	dataptr = databuf;
    }
    return 0;
}


/*
 * Dump output lines are scanned for two types of regex matches.
 *
 * First, there are some cases, unfortunately, where dump detects an
 * error but does not return an error code.  We would like to bring these
 * errors to the attention of the operators anyway.  
 *
 * Second, we attempt to determine what dump thinks its output size is.
 * This is cheaper than putting a filter between dump and compress just
 * to determine the output size.  The re_size table contains regexes to
 * match the size output by various vendors' dump programs.  Some vendors
 * output the number in Kbytes, some in 512-byte blocks.  Whenever an
 * entry in re_size matches, the first integer in the dump line is
 * multiplied by the scale field to get the dump size.
 */

typedef enum { 
    DMP_INFO, DMP_START, DMP_END, DMP_ERROR, DMP_STATS,
    DMP_NORMAL, DMP_STRANGE, DMP_SIZE
} dmpline_t;

typedef struct regex_s {
    dmpline_t typ;
    char *regex;
    int scale;                  /* only used in size table */
} regex_t;

regex_t re_table[] = {
  /* the various encodings of dump size */
  { DMP_SIZE, 
	"DUMP: DUMP: [0-9][0-9]* tape blocks",				1024},
  { DMP_SIZE,
	"dump: Actual: [0-9][0-9]* tape blocks",
                                                                        1024},
  { DMP_SIZE,
        "backup: There are [0-9][0-9]* tape blocks on [0-9]* tapes",    1024},

  { DMP_SIZE,
	"DUMP: [0-9][0-9]* blocks ([0-9][0-9]*KB) on [0-9][0-9]* volume", 512},
  { DMP_SIZE,
"DUMP: [0-9][0-9]* blocks ([0-9][0-9]*\\.[0-9][0-9]*MB) on [0-9][0-9]* volume",
                                                                          512},
  { DMP_SIZE,
	"DUMP: [0-9][0-9]* bytes were dumped",
                                                                            1},

  /* senddump lines */
  { DMP_START,  "^senddump: start .* level [0-9][0-9]*" },
  { DMP_END,    "^senddump: end" },
  { DMP_ERROR,  "^senddump: error" },
  { DMP_STATS,  "^dumpstats: " },

  /* strange dump lines */
  { DMP_STRANGE, "should not happen" },
  { DMP_STRANGE, "Cannot open" },
  /* need to add more ERROR entries here by scanning dump sources */

  /* any blank or non-strange DUMP: lines are marked as normal */
  { DMP_NORMAL, "^  DUMP:" },
  { DMP_NORMAL, "^dump:" },					/* OSF/1 */

  { DMP_NORMAL, "^backup:" },					/* AIX */
  { DMP_NORMAL, "^        Use the umount command to unmount the filesystem" },

  { DMP_NORMAL, "^[ \t]*\\\n" },

  /* catch-all; DMP_STRANGE is returned for all other lines */
  { DMP_STRANGE, NULL, 0}
};

int first_num(str)
char *str;
/*
 * Returns the value of the first integer in a string.
 */
{
    char tmp[16], *tp;

    tp = tmp;
    while(*str && !isdigit(*str)) str++;
    while(*str && isdigit(*str)) *tp++ = *str++;
    *tp = '\0';

    return atoi(tmp);
}

dmpline_t parse_dumpline(str)
char *str;
/*
 * Checks the dump output line in str against the regex table.
 */
{
    regex_t *rp;
    char *errs, *re_comp();

    /* check for error match */
    for(rp = re_table; rp->regex != NULL; rp++) {
        if((errs = re_comp(rp->regex)) != NULL)
	    error("error in regex: %s", errs);
        if(re_exec(str) == 1) /* got match */
	    break;
    }
    if(rp->typ == DMP_SIZE) origsize = first_num(str) * rp->scale;
    return rp->typ;	
}


static char msgbuf[MAX_LINE];
static int msgofs = 0;
enum { 
    START_STATE, NOSIZE_STATE, GOTSIZE_STATE, END_STATE 
} dump_state;

int dump_result;
#define max(a,b) (a>b?a:b)

static void process_dumpeof()
{
    /* process any partial line in msgbuf? !!! */
    if(msgofs != 0) {
	fprintf(errf,"? dumper: error [partial line in msgbuf: %d bytes]\n",
		msgofs);
    }
    switch(dump_state) {	/* these all fall through */
    case START_STATE:
	fputs("? dumper: error [missing start line from senddump]\n",errf);
    case NOSIZE_STATE:
	fputs("? dumper: error [missing size line from dump]\n",errf);
    case GOTSIZE_STATE:
	fputs("? dumper: error [missing end line from senddump]\n",errf);
	dump_result = max(dump_result, 1);
    case END_STATE:
    default:
	/* everthing is fine */;
    }
}

static void process_dumpline(str)
char *str;
{
    char startchr;
    dmpline_t typ;

    typ = parse_dumpline(str);
    if(dump_state == START_STATE && typ != DMP_START) {
	fputs("? dumper: error [missing start line from senddump]\n",errf);
	dump_state = NOSIZE_STATE;
	dump_result = max(dump_result, 1);
    }
    if(dump_state == END_STATE) {
	startchr = '?';
	dump_result = max(dump_result, 1);
    }
    else startchr = '|';

    switch(typ) {
    case DMP_START: 
	if(dump_state == START_STATE) dump_state = NOSIZE_STATE;
	else {
	    startchr = '?';
	    dump_result = max(dump_result, 1);
	}
	break;

    case DMP_STATS:
    case DMP_NORMAL:
	/* regular dump output; do nothing */
	break;
    case DMP_STRANGE: 
	startchr = '?';
	dump_result = max(dump_result, 1);
	break;
    case DMP_SIZE:
	dump_state = GOTSIZE_STATE;
	startchr = '|';
	break;

    case DMP_END:
    case DMP_ERROR:
	if(dump_state != GOTSIZE_STATE) {
	    fputs("? dumper: error [missing size line from dump]\n",errf);
	    startchr = '?';
	    dump_result = max(dump_result, 1);
	}
	dump_state = END_STATE;
	if(typ == DMP_ERROR) {
	    char *p1 = strchr(str,'[');
	    char *p2 = strchr(str,']');
	    if(p1 == NULL) strcpy(errstr,"[unspecified senddump error]");
	    else {
		strcpy(errstr,p1);
		if(p2 != NULL) errstr[p2-p1+1] = '\0';
	    }
	    dump_result = 2;
	}
	break;
    }
    fprintf(errf, "%c %s", startchr, str);
}


static void add_msg_data(str, len)
char *str;
int len;
{
    char *nlpos;
    int len1, got_newline;

    while(len) {

	/* find a newline, if any */
	for(nlpos = str; nlpos < str + len; nlpos++)
	    if(*nlpos == '\n') break;

	/* copy up to newline (or whole string if none) into buffer */
	if(nlpos < str+len && *nlpos == '\n') {
	    got_newline = 1;
	    len1 = nlpos - str + 1;
	}
	else {
	    got_newline = 0;
	    len1 = len;
	}

	/* but don't overwrite the buffer */
	if(len1 + msgofs >= MAX_LINE) {
	    len1 = MAX_LINE-1 - msgofs;
	    str[len1-1] = '\n';			/* force newline */
	    got_newline = 1;
	}
	strncpy(msgbuf + msgofs, str, len1);
	msgofs += len1;
	msgbuf[msgofs] = '\0';

	if(got_newline) {
	    process_dumpline(msgbuf);
	    msgofs = 0;
	}
	len -= len1;
	str += len1;
    }
}


static void log_msgout(typ)
logtype_t typ;
{
    char str[MAX_LINE];

    if((errf = fopen(errfname, "r")) == NULL)
	error("opening msg output: %s", strerror(errno));

    while(fgets(str, MAX_LINE, errf)) {
	str[strlen(str)-1] = '\0';
	log(typ, "%s", str);
    }
    fclose(errf);
}


void update_info()
{
    info_t record;
    stats_t *infp;
    perf_t *perfp;

    /* update info record */

    if(open_infofile(getconf_str(CNF_INFOFILE)))
       error("could not open infofile: %s", strerror(errno));
    get_info(hostname, diskname, &record);
	
    infp = &record.inf[level];
    infp->size = (origsize+1023)/1024;	/* round up */
    infp->csize = (dumpsize+1023)/1024;
    infp->secs = runtime.r.tv_sec + runtime.r.tv_usec / 500000; /* round off */
    infp->date = timestamp;

    if(level == 0) perfp = &record.full;
    else perfp = &record.incr;
    newperf(perfp->comp, origsize? (dumpsize/(float)origsize) : 1.0);
    newperf(perfp->rate, dumpsize/(1024.0*(infp->secs? infp->secs : 1.0)));

    if(interactive)
	printf("dumper: not updating database record for test\n");
    else
	put_info(hostname, diskname, &record);
    close_infofile();
}

void record_success(str)
char *str;
{
    update_info();

    log(L_SUCCESS, "%s %s %d %s", hostname, diskname, level, str);
}

void record_strange(str)
char *str;
{
    update_info();

    log_start_multiline();
    log(L_STRANGE, "%s %s %d %s", hostname, diskname, level, str);
    log_msgout(L_STRANGE);
    log_end_multiline();
}

void record_failure(str)
char *str;
{
    log_start_multiline();
    log(L_FAIL, "%s %s %d %s", hostname, diskname, level, str);
    log_msgout(L_FAIL);
    log_end_multiline();
}


static void do_dump(errfd, infd, outfd)
int errfd, infd, outfd;
{
    int maxfd, nfound, size1, size2, eof1, eof2;
    fd_set readset, selectset;
    struct timeval timeout;

    startclock();
    instr_start();

    dumpsize = origsize = dump_result = msgofs = 0;
    dataptr = databuf;
    spaceleft = DATABUF_SIZE;
    dump_state = START_STATE;

    sprintf(errfname, "/tmp/%s.%s.%d.errout", hostname, diskname, level);
    if((errf = fopen(errfname, "w")) == NULL) {
	sprintf(errstr,"[errfile open %s]", strerror(errno));
	goto failed;
    }

    maxfd = (errfd > infd ? errfd : infd) + 1;
    eof1 = eof2 = 0;

    FD_ZERO(&readset);
    FD_SET(errfd, &readset);

    if(infd == -1) eof1 = 1;	/* fake eof on data */
    else FD_SET(infd, &readset);

    while(!(eof1 && eof2)) {

	timeout.tv_sec = READ_TIMEOUT;
	timeout.tv_usec = 0;
	memcpy(&selectset, &readset, sizeof(fd_set));

	nfound = select(maxfd, &selectset, NULL, NULL, &timeout);

	/* check for errors or timeout */

	if(nfound == 0)  {
	    strcpy(errstr,"[data timeout]");
	    goto failed;
	}
	if(nfound == -1) {
	    sprintf(errstr,  "[select: %s]", strerror(errno));
	    goto failed;
	}

	/* read/write any data */

	if(FD_ISSET(infd, &selectset)) {
	    size1 = read(infd, dataptr, spaceleft);
	    switch(size1) {
	    case -1: 
		sprintf(errstr, "[data read: %s]", strerror(errno));
		goto failed;
	    case 0:
		if(update_dataptr(outfd, size1)) return;
		eof1 = 1;
		close(infd);
		FD_CLR(infd, &readset);
		break;
	    default:
		instr_recvd(size1);
		dumpsize += size1;
		if(update_dataptr(outfd, size1)) return;
	    }
	}
	
	if(FD_ISSET(errfd, &selectset)) {
	    size2 = read(errfd, mesgbuf, MESGBUF_SIZE);
	    switch(size2) {
	    case -1: 
		sprintf(errstr, "[mesg read: %s]", strerror(errno));
		goto failed;
	    case 0:
		eof2 = 1;
		process_dumpeof();
		close(errfd);
		FD_CLR(errfd, &readset);
		break;
	    default:
		add_msg_data(mesgbuf, size2);
	    }
	}
    } /* end while */

    if(dump_result > 1) goto failed;

    runtime = stopclock();

    sprintf(errstr, "[sec %s kb %3.1f kps %3.1f orig-kb %3.1f ]",
	      walltime_str(runtime), dumpsize/1024.0, 
	      dumpsize/(1024*(runtime.r.tv_sec+runtime.r.tv_usec/1000000.0)),
	      origsize/1024.0);
    putresult("DONE %s %s\n", handle, errstr);

    fclose(errf);
    switch(dump_result) {
    case 0:	record_success(errstr);	break;
    case 1:	record_strange(errstr); break;
    }
    unlink(errfname);
    instr_end();
    return;

failed:
    putresult("FAILED %s %s\n", handle, errstr);

    fclose(errf);
    record_failure(errstr);
    unlink(errfname);
}

static int run_dump(hostname, local_port, disk, level, options)
int local_port;
char *hostname, *disk, *level, *options;
{
    int msgfd, len, rc, senddump_port;
    char mesg[256];
    struct servent *senddump;

    if((senddump = getservbyname("amsenddump", "tcp")) == NULL) {
	senddump_port = SENDDUMP_PORT_DEFAULT;
	log(L_WARNING, "no amsenddump/tcp service, using default port");
    }
    else senddump_port = ntohs(senddump->s_port);	/* sigh */

    msgfd = stream_client(hostname, senddump_port, DEFAULT_SIZE, DEFAULT_SIZE);
    if(msgfd < 0) {
	putresult("TRY-AGAIN %s [%s]\n", handle, strerror(errno));
	return -1;
    }

    sprintf(mesg, "%d DUMP %s %s %d %s %s %s\n", local_port, disk, level, 
	    DATABUF_SIZE, datestamp, loginid, options);
    len = strlen(mesg);
    if((rc = write(msgfd, mesg, len)) < len) {
	if(rc >= 0) putresult("FAILED %s [mesg write: short write]\n", handle);
	else putresult("FAILED %s [mesg write: %s]\n", handle, strerror(errno));
	close(msgfd);
	return -1;
    }
    return msgfd;
}

#ifdef INSTRUMENTATION

#define MAX_SECONDS  (4*60*60)	/* up to four hours */
int recvd[MAX_SECONDS+1];
int pckts[MAX_SECONDS+1];
struct timeval start_time, cur_time;
struct timezone dummy_tz;

instr_start()
{
    memset(recvd, 0, sizeof(recvd));
    memset(pckts, 0, sizeof(pckts));
    gettimeofday(&start_time, &dummy_tz);
}

instr_recvd(n)
int n;
{
    int secs;
    gettimeofday(&cur_time, &dummy_tz);
    secs = cur_time.tv_sec - start_time.tv_sec;
    if(secs > MAX_SECONDS) secs = MAX_SECONDS;
    recvd[secs] += n;
    pckts[secs] += 1;
}

instr_end()
{
    int i;
    char filename[256];
    FILE *instrf;

    sprintf(filename, "%s/%s/%s.%s.%d.instr", getconf_str(CNF_DISKDIR),
	    datestamp, hostname, diskname, level);

    if((instrf = fopen(filename, "w")) == NULL)
	error("could not open instr file %s: %s", filename, strerror(errno));

    for(i=0;i<MAX_SECONDS; i++) {
	if(pckts[i]) fprintf(instrf, "%d %d %d\n", i, pckts[i], recvd[i]);
    }
    if(pckts[MAX_SECONDS] != 0) 
	fprintf(instrf, "overflow %d %d\n", 
		pckts[MAX_SECONDS], recvd[MAX_SECONDS]);
    fclose(instrf);
}

#endif /* INSTRUMENTATION */
