/*
 * Amanda, The Advanced Maryland Automatic Network Disk Archiver
 * Copyright (c) 1991,1993 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: James da Silva, Systems Design and Analysis Group
 *			   Computer Science Department
 *			   University of Maryland at College Park
 */
/*
 * planner.c - dump schedule planner for Amanda backup system.
 */
#include "amanda.h"
#include "conffile.h"
#include "diskfile.h"
#include "tapefile.h"
#include "infofile.h"
#include "logfile.h"
#include "clock.h"
#include "dgram.h"

char *pname = "planner";

#define MAX_LEVELS		    3
#define SENDSIZE_PORT_DEFAULT	10069
#define ONE_TIMEOUT		   40	/* # seconds to wait for one est */
#define RETRIES		            3	/* # of retries */
#define PROMOTE_THRESHOLD	 0.05	/* if <5% unbalanced, don't promote */
#define DEFAULT_DUMPRATE         30.0	/* K/s */
#define DEFAULT_COMPRATE         0.50	/* 50% */

#define est_time(d,l)(l?\
		      (est_size(d,l)*compratio(d,incrcomp)/est(d)->incrrate):\
		      (est_size(d,l)*compratio(d,fullcomp)/est(d)->fullrate))
#define compratio(d,v) ((d)->dtype->compress? (est(d)->v) : 1.0)

typedef struct est_s {
    int retries;
    int curr_priority;
    int curr_level;
    int size;
    int degr_level;	/* if curr_level == 0, what would be the inc level */
    int degr_size;
    int last_level;
    int last_lev0size;
    int next_level0;
    int level_days;
    float fullrate, incrrate;
    float fullcomp, incrcomp;
    int level[MAX_LEVELS];
    int est_size[MAX_LEVELS];
} est_t;

#define est(dp)	((est_t *)(dp)->up)

disklist_t waitq, estq, failq, schedq;
long total_size;
double total_lev0, balanced_size, balance_threshold;
unsigned long tape_length, tape_mark;
int result_port, result_socket, sendsize_port;
int total_waiting, max_disks;
char loginid[80];

tapetype_t *tape;

/*
 * ========================================================================
 * MAIN PROGRAM
 *
 */

static void construct_datestamp P((char *buf));		  /* subroutines */
static void setup_estimates P((disklist_t *qp));
static void get_estimates P((void));
static void analyze_estimates P((disklist_t *qp));
static void handle_failed P((disklist_t *qp));
static int delay_lowest_priority_level_zero P((void));
static int delay_lowest_priority_incremental P((void));
static int promote_highest_priority_incremental P((void));
static void output_scheduleline P((disklist_t *qp));

void main()
{
    disklist_t *origqp;
    int moved_one;
    char datestamp[80];
    struct passwd *pwptr;

    erroutput_type = (ERR_AMANDALOG|ERR_INTERACTIVE);
    startclock();


    /*
     * 1. Networking Setup
     *
     * Planner runs setuid to get a priviledged socket for BSD style
     * security.  We get the socket right away as root, then setuid
     * back to a normal user.
     */

    /* set up dgram port first thing */

    if((result_socket = dgram_bind(&result_port)) < 0)
	error("could not bind result datagram port: %s", strerror(errno));

    if(geteuid() == 0) {
	/* set both real and effective uid's to real uid, likewise for gid */
	setgid(getgid());
	setuid(getuid());
    }

    /*
     * From this point on we are running under our real uid, so we don't
     * have to worry about opening security holes below.  Find out who
     * we are running as.
     */

    if((pwptr = getpwuid(getuid())) == NULL)
	error("can't get login name for my uid %d", getuid());
    strcpy(loginid, pwptr->pw_name);


    /*
     * 2. Read in Configuration Information
     *
     * All the Amanda configuration files are loaded before we begin.
     */

    fprintf(stderr,"READING CONF FILES...\n");
     
    if(read_conffile(CONFFILE_NAME))
	error("could not find \"%s\" in this directory.\n", CONFFILE_NAME);

    construct_datestamp(datestamp);
    log(L_START, "date %s", datestamp);

    if((origqp = read_diskfile(getconf_str(CNF_DISKFILE))) == NULL)
	error("could not load \"%s\"\n", getconf_str(CNF_DISKFILE));

    if(read_tapelist(getconf_str(CNF_TAPELIST)))
	error("could not load \"%s\"\n", getconf_str(CNF_TAPELIST));

    if(open_infofile(getconf_str(CNF_INFOFILE)))
	error("could not open info db \"%s\"\n", getconf_str(CNF_INFOFILE));

    tape = lookup_tapetype(getconf_str(CNF_TAPETYPE));
    tape_length = tape->length;
    tape_mark   = tape->filemark;

    fprintf(stderr, "startup took %s secs\n", walltime_str(curclock()));

    /*
     * 3. Calculate Preliminary Dump Levels
     *
     * Before we can get estimates from the remote slave hosts, we make a
     * first attempt at guessing what dump levels we will be dumping at
     * based on the curinfo database.
     */

    fprintf(stderr,"\nSETTING UP FOR ESTIMATES...\n");
    startclock();
     
    waitq.head = waitq.tail = NULL;
    while(!empty(*origqp)) setup_estimates(origqp);

    fprintf(stderr, "setting up estimates took %s secs\n",
	    walltime_str(curclock()));


    /*
     * 4. Get Dump Size Estimates from Remote Slave Hosts
     *
     * Each host is queried (in parallel) for dump size information on all
     * of its disks, and the results gathered as they come in.
     */

    /* go out and get the dump estimates */

    fprintf(stderr,"\nGETTING ESTIMATES...\n");

    estq.head = estq.tail = NULL;
    failq.head = failq.tail = NULL;

    get_estimates();

    /*
     * At this point, all disks with estimates are in estq, and
     * all the disks on hosts that didn't respond to our inquiry
     * are in failq.
     */

    dump_queue("FAILED", failq, 15, stderr);
    dump_queue("DONE", estq, 15, stderr);


    /*
     * 5. Analyze Dump Estimates
     *
     * Each disk's estimates are looked at to determine what level it
     * should dump at, and to calculate the expected size and time taking
     * historical dump rates and compression ratios into account.  The
     * total expected size is accumulated as well.
     */

    fprintf(stderr,"\nANALYZING ESTIMATES...\n");

    startclock();

    schedq.head = schedq.tail = NULL;
    while(!empty(estq)) analyze_estimates(&estq);
    while(!empty(failq)) handle_failed(&failq);

    /*
     * At this point, all the disks are on schedq sorted by priority.
     * The total estimated size of the backups is in total_size.
     */

    {
	int pos;
	disk_t *dp;

	fprintf(stderr, "INITIAL SCHEDULE:\n");
	for(pos = 0, dp = schedq.head; dp != NULL; dp = dp->next, pos++) {
	    fprintf(stderr, "  %s %s pri %d lev %d size %d\n",
		    dp->host->hostname, dp->name, est(dp)->curr_priority,
		    est(dp)->curr_level, est(dp)->size);
	}
    }


    /*
     * 6. Delay Dumps if Schedule Too Big
     *
     * If the generated schedule is too big to fit on the tape, we need to
     * delay some full dumps to make room.  Incrementals will be done
     * instead (except for new or forced disks).
     *
     * In extreme cases, delaying all the full dumps is not even enough.
     * If so, some low-priority incrementals will be skipped completely
     * until the dumps fit on the tape.
     */

    fprintf(stderr,
	 "\nDELAYING DUMPS IF NEEDED, total_size %d, tape length %u mark %u\n",
	    total_size, tape_length, tape_mark);

    moved_one = 1;
    while(total_size > tape_length && moved_one)
	moved_one = delay_lowest_priority_level_zero();

    moved_one = 1;
    while(total_size > tape_length && moved_one)
	moved_one = delay_lowest_priority_incremental();

    if(total_size <= 0)
	error("cannot fit anything on tape, bailing out\n");


    /*
     * 7. Promote Dumps if Schedule Too Small
     *
     * Amanda attempts to balance the full dumps over the length of the
     * dump cycle.  If this night's full dumps are too small relative to
     * the other nights, promote some high-priority full dumps that will be
     * due for the next run, to full dumps for tonight, taking care not to
     * overflow the tape size.
     *
     * Amanda never delays full dumps just for the sake of balancing the
     * schedule, so it can take a full cycle to balance the schedule after
     * a big bump.
     */

    fprintf(stderr,
     "\nPROMOTING DUMPS IF NEEDED, total_lev0 %1.2f, balanced_size %1.2f...\n",
	    total_lev0, balanced_size);

    balance_threshold = balanced_size * PROMOTE_THRESHOLD;
    moved_one = 1;
    while((balanced_size - total_lev0) > balance_threshold && moved_one)
	moved_one = promote_highest_priority_incremental();

    fprintf(stderr, "analysis took %s secs\n", walltime_str(curclock()));


    /*
     * 8. Output Schedule
     *
     * The schedule goes to stdout, presumably to driver.  A copy is written
     * on stderr for the debug file.
     */

    fprintf(stderr,"\nGENERATING SCHEDULE:\n");

    while(!empty(schedq)) output_scheduleline(&schedq);
    log(L_FINISH, "date %s", datestamp);
}

static void construct_datestamp(buf)
char *buf;
{
    struct tm *tm;
    time_t timevar;

    timevar = time((time_t *)NULL);
    tm = localtime(&timevar);
    sprintf(buf, "%04d%02d%02d", tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday);
}




/*
 * ========================================================================
 * SETUP FOR ESTIMATES
 *
 */

static int last_level P((disk_t *dp, info_t *ip));	  /* subroutines */
static int est_size P((disk_t *dp, int level));
static int failed_lastnight P((disk_t *dp, info_t *ip));
static int next_level0 P((disk_t *dp, info_t *ip));
static int days_at P((disk_t *dp, info_t *ip, int lev));
static int bump_thresh P((int level));
static int time_to_datestamp P((time_t timevar));


static void setup_estimates(qp)
disklist_t *qp;
{
    disk_t *dp;
    est_t *ep;
    info_t inf;
    stats_t *s;
    int i, curr_level;

    dp = dequeue_disk(qp);

    assert(dp && dp->dtype && dp->host);

    /* get current information about disk */

    if(get_info(dp->host->hostname, dp->name, &inf)) {
	/* no record for this disk, make a note of it */
	log(L_INFO, "Adding new disk %s:%s.", dp->host->hostname, dp->name);
    }

    /* setup working data struct for disk */

    ep = alloc(sizeof(est_t));
    dp->up = (void *) ep;
    ep->retries = 0;
    ep->size = -1;
    ep->curr_priority = dp->dtype->priority;

    /* calculated fields */

    if(inf.command == PLANNER_FORCE) {
	/* force a level 0, kind of like a new disk */
	if(dp->dtype->no_full) {
	    /* 
	     * XXX - Not sure what it means to force a no-full disk.  The
	     * purpose of no-full is to just dump changes relative to a
	     * stable base, for example root partitions that vary only
	     * slightly from a site-wide prototype.  Only the variations
	     * are dumped.
	     *
	     * If we allow a level 0 onto the Amanda cycle, then we are
	     * hosed when that tape gets re-used next.  Disallow this for
	     * now.
	     */
	    log(L_ERROR,
		"Cannot force full dump of %s:%s with no-full option.",
		dp->host->hostname, dp->name);

	    /* clear force command */
	    if(inf.command == PLANNER_FORCE)
		inf.command = NO_COMMAND;
	    if(put_info(dp->host->hostname, dp->name, &inf))
		error("could not put info record for %s:%s: %s",
		      dp->host->hostname, dp->name, strerror(errno));
	    ep->last_level = last_level(dp, &inf);
	    ep->next_level0 = next_level0(dp, &inf);
	}
	else {
	    ep->last_level = -1;
	    ep->next_level0 = 0;
	    log(L_INFO, "Forcing full dump of %s:%s as directed.",
		dp->host->hostname, dp->name);
	}
    }
    else {
	ep->last_level = last_level(dp, &inf);
	ep->next_level0 = next_level0(dp, &inf);
    }

    /* adjust priority levels */

    if(ep->next_level0 < 0) {
	fprintf(stderr,"%s:%s overdue %d days for level 0\n",
		dp->host->hostname, dp->name, - ep->next_level0);
	ep->curr_priority -= ep->next_level0;
    }
    else if(failed_lastnight(dp, &inf) || inf.command == PLANNER_FORCE) {
	ep->curr_priority += 1;
    }

    /* handle external level 0 dumps */

    if(dp->dtype->skip_full) {
	if(ep->next_level0 <= 0) {
	    log(L_INFO, "Skipped full dump of %s:%s on this amdump run.",
		dp->host->hostname, dp->name);
	    /* update the date field */
	    time(&inf.inf[0].date);
	    /* clear force command */
	    if(inf.command == PLANNER_FORCE)
		inf.command = NO_COMMAND;
	    if(put_info(dp->host->hostname, dp->name, &inf))
		error("could not put info record for %s:%s: %s",
		      dp->host->hostname, dp->name, strerror(errno));
	    fprintf(stderr,"%s:%s lev 0 skipped due to skip-full flag\n",
		    dp->host->hostname, dp->name);
	    /* don't enqueue the disk */
	    ep->level[0] = ep->level[1] = ep->level[2] = -1;
	    log(L_SUCCESS, "%s %s 0 [skipped: skip-full]", 
		dp->host->hostname, dp->name);
	    return;
	}
	else if(ep->next_level0 == 1) {
	    log(L_WARNING, "Skipping full dump of %s:%s on next amdump run.",
		dp->host->hostname, dp->name);
	}
    }

    /* handle "skip-incr" type archives */

    if(dp->dtype->skip_incr && ep->next_level0 > 0) {
	fprintf(stderr,"%s:%s lev 1 skipped due to skip-incr flag\n",
		dp->host->hostname, dp->name);
	/* don't enqueue the disk */
	ep->level[0] = ep->level[1] = ep->level[2] = -1;
	log(L_SUCCESS, "%s %s 1 [skipped: skip-incr]",
	    dp->host->hostname, dp->name);
	return;
    }

    if(ep->last_level == 0) ep->level_days = 0;
    else ep->level_days = days_at(dp, &inf, ep->last_level);
    ep->last_lev0size = inf.inf[0].csize;

    ep->fullrate = perf_average(inf.full.rate, DEFAULT_DUMPRATE);
    ep->incrrate = perf_average(inf.incr.rate, DEFAULT_DUMPRATE);
    ep->fullcomp = perf_average(inf.full.comp, DEFAULT_COMPRATE);
    ep->incrcomp = perf_average(inf.incr.comp, DEFAULT_COMPRATE);

    /* determine which estimates to get */

    i = 0;
    if(!(dp->dtype->skip_full || dp->dtype->no_full))
	ep->level[i++] = 0;
    if(ep->last_level == -1) {		/* a new disk */
	if(dp->dtype->no_full)
	    ep->level[i++] = 1;
	else assert(!dp->dtype->skip_full);	/* should be handled above */
    }
    else {				/* not new, pick normally */
	curr_level = ep->last_level;
	if(curr_level == 0)
	    ep->level[i++] = 1;
	else {
	    ep->level[i++] = curr_level;
	    /*
	     * If last time we dumped less than the threshold, then this
	     * time we will too, OR the extra size will be charged to both
	     * cur_level and cur_level + 1, so we will never bump.  Also,
	     * if we haven't been at this level 2 days, or the dump failed
	     * last night, we can't bump.
	     */
	    if((inf.inf[curr_level].size == 0 || /* no data, try it anyway */
		(inf.inf[curr_level].size > bump_thresh(curr_level))) &&
		ep->level_days >= getconf_int(CNF_BUMPDAYS) &&
		!failed_lastnight(dp, &inf))
		ep->level[i++] = curr_level+1;
	}
    }
    while(i < MAX_LEVELS) {	/* mark end of estimates */
	ep->level[i] = -1;
	ep->est_size[i] = -1;
	i++;
    }

    /* For each estimate, start with the historical size, if any */

    for(i=0; i < MAX_LEVELS; i++) {
	if(ep->level[i] == -1) break;
	s = &inf.inf[ep->level[i]];
	if(s->date == EPOCH) ep->est_size[i] = -1;
	else ep->est_size[i] = s->size;
    }

    /* debug output */

    fprintf(stderr, "setup_estimates: %s:%s: command %d, options:",
	    dp->host->hostname, dp->name, inf.command);
    if(dp->dtype->no_full) fputs(" no-full", stderr);
    if(dp->dtype->skip_full) fputs(" skip-full", stderr);
    if(dp->dtype->skip_incr) fputs(" skip-incr", stderr);
    fprintf(stderr, "\n    last_level %d next_level0 %d level_days %d\n",
	    ep->last_level, ep->next_level0, ep->level_days);
    fprintf(stderr, "    getting estimates %d (%d) %d (%d) %d (%d)\n",
	    ep->level[0], ep->est_size[0], 
	    ep->level[1], ep->est_size[1],
	    ep->level[2], ep->est_size[2]);

    assert(ep->level[0] != -1);
    enqueue_disk(&waitq, dp);
}

/* calculations based on tapelist and current disk info */

/* what was the level of the last successful dump to tape? */
static int last_level(dp, ip)
disk_t *dp;
info_t *ip;
{
    int min_pos, min_level, i;
    time_t lev0_date;
    tape_t *tp;

    min_pos = 99;
    min_level = -1;
    lev0_date = 0;
    for(i = 0; i < 9; i++) {
	if((tp = lookup_tapelabel(ip->inf[i].label)) == NULL) continue;
	/* cull any entries from previous cycles */
	if(i == 0) lev0_date = ip->inf[0].date;
	else if(ip->inf[i].date < lev0_date) continue;

	if(tp->position < min_pos) {
	    min_pos = tp->position;
	    min_level = i;
	}
    }
    return min_level;
}
 
static int est_size(dp, level)
disk_t *dp;
int level;
{
    int i;

    for(i = 0; i < MAX_LEVELS; i++) {
	if(level == est(dp)->level[i])
	    return est(dp)->est_size[i];
    }
    return -1;
}


static int failed_lastnight(dp, ip)	/* did it fail last night? */
disk_t *dp;
info_t *ip;
{
    tape_t *last_tape = lookup_tapepos(1);	/* last night's tape */
    int last_dump = last_level(dp, ip);

    if(last_tape == NULL || last_dump == -1 ||
       strcmp(ip->inf[last_dump].label, last_tape->label))
	return 1;
    else
	return 0;
}

/* when is next level 0 due? 0 = tonight, 1 = tommorrow, etc*/
static int next_level0(dp, ip)
disk_t *dp;
info_t *ip;
{
    int duedays;
    tape_t *lev0_tape;

    if(dp->dtype->no_full) return 1;	/* fake it */

    if(dp->dtype->skip_full)
	lev0_tape = lookup_tapedate(time_to_datestamp(ip->inf[0].date));
    else
	lev0_tape = lookup_tapelabel(ip->inf[0].label);

    if(lev0_tape == NULL) return 0;

    duedays = dp->dtype->mincycle - lev0_tape->position;
    return duedays;
}

/* how many days at level? */
static int days_at(dp, ip, lev)
disk_t *dp;
info_t *ip;
int lev;
{
    tape_t *cur_tape, *old_tape;
    int last;

    last = last_level(dp, ip);
    if(lev != last) return 0;
    if(lev == 0) return 1;

    cur_tape = lookup_tapelabel(ip->inf[lev].label);
    old_tape = lookup_tapelabel(ip->inf[lev-1].label);
    if(cur_tape == NULL || old_tape == NULL) return 0;

    return old_tape->position - cur_tape->position;
}


static int bump_thresh(level)
int level;
{
    int bump = getconf_int(CNF_BUMPSIZE);
    double mult = getconf_real(CNF_BUMPMULT);

    while(--level) bump = (int) bump * mult;
    return bump;
}

static int time_to_datestamp(timevar)
time_t timevar;
{
    struct tm *tm;

    tm = localtime(&timevar);
    return (tm->tm_year+1900)*10000 + (tm->tm_mon+1)*100 + tm->tm_mday;
}




/*
 * ========================================================================
 * GET REMOTE DUMP SIZE ESTIMATES
 *
 */

static host_t *handle2host P((char *str));		  /* subroutines */
static char *host2handle P((host_t *hostp));
static void getsize P((disklist_t *qp, host_t *hostp));
static void reap_results_til_timeout P((void));
static disk_t *lookup_hostdisk P((host_t *hp, char *str));
static int handle_result P((dgram_t *msg));


static void get_estimates()
{
    disklist_t processed;
    struct servent *sendsize;
    int tries_left;

    if((sendsize = getservbyname("amsendsize", "udp")) == NULL)
	sendsize_port = SENDSIZE_PORT_DEFAULT;
    else
	sendsize_port = ntohs(sendsize->s_port);	/* sigh */

    tries_left = RETRIES;
    while(tries_left--) {

      total_waiting = 0;
      max_disks = 0;

      fprintf(stderr, "sending dgrams, try # %d\n", RETRIES-tries_left);
      startclock();

      processed.head = processed.tail = NULL;
      while(!empty(waitq)) {
	getsize(&processed, waitq.head->host);
      }
      waitq = processed;

      fprintf(stderr, "sending dgrams took %s secs\n", 
	      walltime_str(curclock()));

      reap_results_til_timeout();
    }
    failq = waitq;
}

static host_t *handle2host(str)
char *str;
{
    int i;

    sscanf(str, "%X", &i);
    return (host_t *) i;
}

static char *host2handle(hostp)
host_t *hostp;
{
    static char str[80];

    sprintf(str, "%X", (int) hostp);
    return str;
}


static void getsize(qp, hostp)
disklist_t *qp;
host_t *hostp;
{
    dgram_t *msg;
    disk_t *dp;
    char line[1024], myhostname[256];
    int i, disks;

    msg = dgram_alloc();

    gethostname(myhostname, 256);

#ifdef notyet
    sprintf(line, "dumpuser %s\n", loginid);
#endif
    sprintf(line, "replyhost %s\n", myhostname);
    dgram_cat(msg, line);
    sprintf(line, "replyport %d\n", result_port);
    dgram_cat(msg, line);
    dgram_cat(msg, "program DUMP\n");
    sprintf(line, "handle %s\n", host2handle(hostp));
    dgram_cat(msg, line);

    disks = 0;
    for(dp = hostp->disks; dp != NULL; dp = dp->hostnext) {
	if(est(dp)->level[0] == -1) continue;	/* ignore this disk */

	remove_disk(&waitq, dp);
	enqueue_disk(qp, dp);

	sprintf(line, "%s ", dp->name);
	dgram_cat(msg, line);
	for(i=0;i<MAX_LEVELS;i++) {
	    if(est(dp)->level[i] == -1) break;
	    disks++;
	    sprintf(line, "%d", est(dp)->level[i]);
	    dgram_cat(msg, line);
	}
	dgram_cat(msg, "\n");
    }
    hostp->busy = 0;	/* XXX workaround for problem elsewhere? */
    if(disks > max_disks) max_disks = disks;

    fprintf(stderr,"sending dgram to %s, handle %s:\n", hostp->hostname, 
	   host2handle(hostp));

    if(dgram_send(hostp->hostname, sendsize_port, msg))
	error("could not send request to %s\n", hostp->hostname);
    total_waiting ++;
}

static void reap_results_til_timeout()
{
    dgram_t *inmsg;
    int rc, timeout;
    struct timeval starttime, curtime, stoptime;
    struct timezone dont_care;

    inmsg = dgram_alloc();

    /* determine timeout */
    gettimeofday(&starttime, &dont_care);
    stoptime = starttime;
    stoptime.tv_sec += ONE_TIMEOUT * max_disks;

    fprintf(stderr,"entering wait loop, max_disks = %d\n", max_disks);

    while(total_waiting) {
	gettimeofday(&curtime, &dont_care);
	timeout = stoptime.tv_sec - curtime.tv_sec;

	fprintf(stderr,"waiting %d secs for %d results...\n",
		timeout, total_waiting);
	fflush(stderr);

	rc = dgram_recv(result_socket, inmsg, timeout, NULL);
	if(rc == 0) return;	/* timeout */
	if(rc < 0) error("dgram_recv");
	if(handle_result(inmsg)) total_waiting--;
    }
    gettimeofday(&curtime, &dont_care);
    fprintf(stderr,"Got all results after %d seconds\n",
	    curtime.tv_sec - starttime.tv_sec);
}

static disk_t *lookup_hostdisk(hp, str)
host_t *hp;
char *str;
{
    disk_t *dp;

    for(dp = hp->disks; dp != NULL; dp = dp->hostnext)
	if(!strcmp(str, dp->name)) return dp;

    return NULL;
}


static int handle_result(msg)
dgram_t *msg;
{
    char msghost[80], handle[80], msgdisk[80];
    int err, i, rc, norm_speed, comp_speed, size[MAX_LEVELS];
    disk_t *dp;
    host_t *hp;

    err = 1;
    if((rc = sscanf(msg->cur, "hostname %s", msghost)) < 1) goto bad_msg;
    dgram_eatline(msg);

    err = 2;
    if((rc = sscanf(msg->cur, "handle %s", handle)) < 1) goto bad_msg;
    dgram_eatline(msg);
    err = 3;
    if((hp = handle2host(handle)) == NULL) {
	fprintf(stderr, "stale handle from %s (handle %s)\n", msghost, handle);
	goto bad_msg;
    }
    err = 4;
    if(hp->busy) {
	fprintf(stderr, "duplicate message from %s (handle %s)\n",
		hp->hostname, handle);
	goto bad_msg;
    }

    hp->busy = 1;	/* got result for host */

    err = 5;
    if((rc = sscanf(msg->cur, "speed %d %d", &norm_speed, &comp_speed)) < 1)
	goto bad_msg;
    dgram_eatline(msg);

    while(*msg->cur) {
	err++;
	assert(MAX_LEVELS == 3);	/* == the number of %d's below */

	rc = sscanf(msg->cur, "%s %d %d %d", msgdisk,
		    &size[0], &size[1], &size[2]);

	dgram_eatline(msg);
	if(rc < 1) goto bad_msg;
	dp = lookup_hostdisk(hp, msgdisk);
	
	err++;
	for(i = 0; i < MAX_LEVELS; i++) {
	    if(est(dp)->level[i] < 0) break;
	    est(dp)->est_size[i] = size[i];
	}
	if(rc < i+1) goto bad_msg;

	remove_disk(&waitq, dp);
	enqueue_disk(&estq, dp);
	fprintf(stderr,"got result for host %s disk %s:", 
	       dp->host->hostname, dp->name);
	fprintf(stderr," %d -> %dK, %d -> %dK, %d -> %dK\n",
	       est(dp)->level[0], est(dp)->est_size[0],
	       est(dp)->level[1], est(dp)->est_size[1],
	       est(dp)->level[2], est(dp)->est_size[2]);
    }
fflush(stderr);
    return 1;
bad_msg:
    fprintf(stderr,"got a bad message, err = %d:\n", err);
    fprintf(stderr,"----\n%s----\n", msg->data);
    fflush(stderr);
    return 0;
}




/*
 * ========================================================================
 * ANALYSE ESTIMATES
 *
 */

static int schedule_order P((disk_t *a, disk_t *b));	  /* subroutines */
static int pick_inclevel P((disk_t *dp));

static void analyze_estimates(qp)
disklist_t *qp;
{
    int lev0size;
    disk_t *dp;

    dp = dequeue_disk(qp);

    fprintf(stderr,"pondering %s:%s... ",
	    dp->host->hostname, dp->name);
    fprintf(stderr,"next_level0 %d last_level %d ",
	    est(dp)->next_level0, est(dp)->last_level);

    if(est(dp)->next_level0 <= 0) {
	fprintf(stderr,"(due for level 0) ");
	est(dp)->curr_level = 0;
	est(dp)->size = est_size(dp, 0) * compratio(dp,fullcomp);
	total_lev0 += (double) est(dp)->size;
	if(est(dp)->last_level == -1 || dp->dtype->skip_incr) {
	    fprintf(stderr,"(%s disk, can't switch to degraded mode)\n",
		    dp->dtype->skip_incr? "skip-incr":"new");
	    est(dp)->degr_level = -1;
	    est(dp)->degr_size = -1;
	}
	else {
	    /* fill in degraded mode info */
	    fprintf(stderr,"(picking inclevel for degraded mode)\n");
	    est(dp)->degr_level = pick_inclevel(dp);
	    est(dp)->degr_size = 
		est_size(dp, est(dp)->degr_level) * compratio(dp,incrcomp);
	}
    }
    else {
	fprintf(stderr,"(not due for a full dump, picking an incr level)\n");
	est(dp)->curr_level = pick_inclevel(dp);
	est(dp)->size = 
	    est_size(dp,est(dp)->curr_level) * compratio(dp,incrcomp);
    }

    fprintf(stderr,"  curr level %d size %d ", est(dp)->curr_level,
	    est(dp)->size);

    total_size += est(dp)->size + tape_mark;

    if(!(dp->dtype->skip_full || dp->dtype->no_full)) {
	/* calculate level 0 size for balancing */
	if(est_size(dp, 0) == -1) lev0size = est(dp)->last_lev0size;
	else lev0size = est_size(dp, 0) * compratio(dp,fullcomp);
	balanced_size += (lev0size / dp->dtype->mincycle);
    }
    insert_disk(&schedq, dp, schedule_order);

    fprintf(stderr,"total size %d total_lev0 %1.2f balanced-lev0size %1.2f\n",
	    total_size, total_lev0, balanced_size);
}

static void handle_failed(qp)
disklist_t *qp;
{
    disk_t *dp;

    dp = qp->head;

    log(L_WARNING, "Could not get estimate for %s:%s, using historical data.",
	   dp->host->hostname, dp->name);
    analyze_estimates(qp);
}


static int schedule_order(a, b)
disk_t *a, *b;
/*
 * insert-sort by decreasing priority, then
 * by increasing size within priority levels.
 */
{
    int diff;

    if((diff = est(b)->curr_priority - est(a)->curr_priority) != 0)
	return diff;
    else
	return est(a)->size - est(b)->size;
}


static int pick_inclevel(dp)
disk_t *dp;
{
    int base_level, size1, size2;

    base_level = est(dp)->last_level;

    /* if last night was level 0, do level 1 tonight, no ifs ors buts */
    if(base_level == 0) {
	fprintf(stderr,"   picklev: last night 0, so tonight level 1\n");
	return 1;
    }
    /* if no-full option set, always do level 1 */
    if(dp->dtype->no_full) {
	fprintf(stderr,"   picklev: no-full set, so always level 1\n");
	return 1;
    }

    size1 = (int) est_size(dp, base_level);
    fprintf(stderr,
	    "   pick: orig sz %d comp %1.3f out sz %d rate %1.3f time %d\n",
	    size1, compratio(dp,incrcomp),
	    (int)(est_size(dp, base_level)* compratio(dp,incrcomp)),
	    est(dp)->incrrate, (int)est_time(dp, base_level));

    fprintf(stderr,
    "   pick: base size %d base level %d days %d (thresh %dK, %d days)\n",
	    size1, base_level, est(dp)->level_days, 
	    bump_thresh(base_level), getconf_int(CNF_BUMPDAYS));

    if(base_level != 9 && est(dp)->level_days >= getconf_int(CNF_BUMPDAYS)
       && est_size(dp, base_level+1) != -1 
       && size1 > bump_thresh(base_level)) {

	size2 = (int) est_size(dp, base_level+1);
	fprintf(stderr,"   picklev: next size %d... ", size2);
	if(size1 - size2 > bump_thresh(base_level)) {
	    base_level++;
	    fputs("BUMPED\n",stderr);
	    log(L_INFO, "Incremental of %s:%s bumped to level %d.",
		dp->host->hostname, dp->name, base_level);
	}
	else {
	    fputs("not bumped\n",stderr);
	}
    }
    return base_level;
}




/*
 * ========================================================================
 * ADJUST SCHEDULE
 *
 */

static int delay_lowest_priority_level_zero() 
/* try to move a level 0 to tommorrow */
{
    disk_t *ptr;

    for(ptr = schedq.tail; ptr != NULL; ptr = ptr->prev) {
	if(est(ptr)->curr_level == 0) {
	    total_size -= est(ptr)->size;
	    total_lev0 -= est(ptr)->size;

	    if(est(ptr)->last_level == -1 || ptr->dtype->skip_incr) {
		remove_disk(&schedq, ptr);
		fprintf(stderr,
			"  delay: zapping %s disk %s:%s, total size now %d\n",
			ptr->dtype->skip_incr? "skip-incr": "new",
			ptr->host->hostname, ptr->name, total_size);
		log(L_FAIL, 
		"%s %s 0 [dumps too big, but cannot incremental dump %s disk]",
		    ptr->host->hostname, ptr->name,
		    ptr->dtype->skip_incr? "skip-incr": "new");

		/* dump totally gone, forget the filemark too */

		total_size -= tape_mark;
		total_lev0 -= tape_mark;
		return 1;
	    }
	    else {
		est(ptr)->curr_level = est(ptr)->degr_level;
		est(ptr)->size = est(ptr)->degr_size;
		total_size += est(ptr)->size;
		fprintf(stderr,
		      "  delay: moving %s:%s to level %d, total size now %d\n",
			ptr->host->hostname, ptr->name, est(ptr)->curr_level, 
			total_size);
		log(L_INFO,
		    "Dumps too big for tape: full dump of %s:%s delayed.",
		    ptr->host->hostname, ptr->name);
		return 1;
	    }
	}
    }
    return 0;
}


static int delay_lowest_priority_incremental() 
/* try to move an incremental to tommorrow */
{
    disk_t *ptr;

    for(ptr = schedq.tail; ptr != NULL; ptr = ptr->prev) {
	if(est(ptr)->curr_level != 0) {
	    total_size -= est(ptr)->size + tape_mark;

	    remove_disk(&schedq, ptr);
	    fprintf(stderr,
		    "  delay: zapping disk %s:%s lev %d, total size now %d\n",
		    ptr->host->hostname, ptr->name, 
		    est(ptr)->curr_level, total_size);
	    log(L_FAIL, 
		"%s %s %d [dumps way too big, must skip incremental dumps]",
		    ptr->host->hostname, ptr->name, est(ptr)->curr_level);

	    return 1;
	}
    }

    return 0;
}


static int promote_highest_priority_incremental()
{
    disk_t *ptr;
    int new_size, new_total, check_days, check_limit;

    /*
     * return 1 if did so; must update total_size correctly; must not
     * cause total_size to exceed tape_length
     */
    check_limit = getconf_int(CNF_MINCYCLE)-1;
    fprintf(stderr,"   promote: checking up to %d days ahead\n", 
	    check_limit-1);

    for(check_days = 1; check_days < check_limit; check_days++) {
	fprintf(stderr,"   promote: checking %d days now\n", check_days);

        for(ptr = schedq.head; ptr != NULL; ptr = ptr->next) {
	    if(ptr->dtype->skip_full || ptr->dtype->no_full) {
		fprintf(stderr,
		     "    promote: can't move %s:%s: no full dumps allowed.\n",
			ptr->host->hostname, ptr->name);
                continue;
            }
	    if(est(ptr)->next_level0 == check_days) {
		new_size = (int) (est_size(ptr, 0) * compratio(ptr,fullcomp));
		new_total = total_size - est(ptr)->size + new_size;
		if(new_total <= tape_length && 
		   total_lev0 + new_size <=
		   balanced_size + balance_threshold) {
		    total_size = new_total;
		    total_lev0 += new_size;
		    est(ptr)->degr_level = est(ptr)->curr_level;
		    est(ptr)->degr_size = est(ptr)->size;
		    est(ptr)->curr_level = 0;
		    est(ptr)->next_level0 = 0;
		    est(ptr)->size = new_size;

		    fprintf(stderr,
	      "   promote: moving %s:%s up, total_lev0 %1.2f, total_size %d\n",
			    ptr->host->hostname, ptr->name, 
			    total_lev0, total_size);

		    log(L_INFO,
			"Full dump of %s:%s promoted from %d days ahead.",
			ptr->host->hostname, ptr->name, check_days);
		    return 1;
		}
		else {
		    fprintf(stderr,
			    "  promote: %s:%s too big: new size %d total %1.1f, bal size %1.1f thresh %1.1f\n",
			    ptr->host->hostname, ptr->name, new_size, 
			    total_lev0+new_size, balanced_size, 
			    balance_threshold);
		}
	    }
	}
    }
    return 0;
}




/*
 * ========================================================================
 * OUTPUT SCHEDULE
 *
 */
static void output_scheduleline(qp)
disklist_t *qp;
{
    disk_t *dp;
    int time, degr_time;
    char schedline[1024];

    dp = dequeue_disk(qp);
    if(est(dp)->size == -1) {
	/* no estimate, fail the disk */
	log(L_FAIL, "%s %s %d [no estimate or historical data]", 
	    dp->host->hostname, dp->name, est(dp)->curr_level);
	return;
    }
    else if(est(dp)->curr_level == 0 && est(dp)->degr_level != -1) {
	time = est(dp)->size / est(dp)->fullrate;
	degr_time = est(dp)->degr_size / est(dp)->incrrate;
	sprintf(schedline, "%s %s %d %d %d %d %d %d %d\n",
		dp->host->hostname, dp->name, est(dp)->curr_priority,
		est(dp)->curr_level, est(dp)->size, time,
		est(dp)->degr_level, est(dp)->degr_size, degr_time);
    }
    else {
	time = est(dp)->size / est(dp)->incrrate;
	sprintf(schedline, "%s %s %d %d %d %d\n",
		dp->host->hostname, dp->name, est(dp)->curr_priority,
		est(dp)->curr_level, est(dp)->size, time);
    }
    fputs(schedline, stdout);
    fputs(schedline, stderr);
}
