/*
 * Amanda, The Advanced Maryland Automatic Network Disk Archiver
 * Copyright (c) 1991 University of Maryland
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.M. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.M. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * U.M. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL U.M.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: James da Silva, Systems Design and Analysis Group
 *			   Computer Science Department
 *			   University of Maryland at College Park
 */
/* 
 * sendsize.c - send estimated dump sizes
 */
#include "amanda.h"
#include "dgram.h"

#ifdef SYSV_SETPGRP
#  define SETPGRP	setpgrp()
#else
#  define SETPGRP	setpgrp(0,getpid())
#endif

#define DEBUG			/* write debug output to DEBUG_OUTPUT */

#define DEBUG_OUTPUT "/tmp/sendsize.out"
#define MAXLINE 4096
#define SPEEDS "50 60"

#ifdef DEBUG
#  define debug_write(s)	write(1,(s),strlen(s))
#else
#  define debug_write(s)
#endif

typedef struct regex_s {
    char *regex;
    int scale;
} regex_t;

regex_t re_size[] = {
    {"  DUMP: estimated [0-9][0-9]* tape blocks", 1024},
    {"  DUMP: estimated [0-9][0-9]* blocks", 512},
    {"  DUMP: Estimated [0-9][0-9]* bytes", 1},
    {"dump: Estimate: [0-9][0-9]* tape blocks", 1024},		    /* OSF/1 */
    {"backup: There are an estimated [0-9][0-9]* tape blocks.",1024}, /* AIX */

    { NULL, 0 }
};


dgram_t *inmsg, *outmsg;
char line[MAXLINE];

char *pname = "sendsize";

/* local functions */
static int getsize P((char *disk, int level));
static int handle_dumpline P((char *str));

void main(argc, argv)
int argc;
char **argv;
{
    int replyport, size, db, myuid, userok;
    time_t curtime;
    char myhost[256], program[256], handle[256];
    char disk[16], levels[40], *lp, msg[80], *ctime();
    char remotehost[80], remoteuser[80], localuser[80], *pwname;
    struct passwd *pwptr;
    struct sockaddr_in peername;
    struct hostent *hp, *gethostbyaddr();

    erroutput_type = (ERR_INTERACTIVE|ERR_SYSLOG);

#ifdef FORCE_DUMPUSER

    /* we'd rather not run as root */

    if(geteuid() == 0) {
	pwname = SLAVE_LOGIN;
	if((pwptr = getpwnam(pwname)) == NULL)
	    error("cannot find user %s in passwd file.\n", pwname);

	initgroups(pwname, pwptr->pw_gid);
	setgid(pwptr->pw_gid);
	setuid(pwptr->pw_uid);
    }

#endif

    /* setup debug output */

    chdir("/tmp");	/* somewhere we can dump core :-) */

    umask(0);
    if((db = open(DEBUG_OUTPUT, O_WRONLY|O_TRUNC|O_CREAT, 0666)) == -1)
	db = open("/dev/null", O_WRONLY);

    dup2(db, 1);
    dup2(db, 2);

    time(&curtime);
    sprintf(msg, "sendsize: called @ %s-----\n", ctime(&curtime));
    debug_write(msg);

    /* receive message */
    
    inmsg = dgram_alloc();
    outmsg = dgram_alloc();
    if(dgram_recv(0, inmsg, 10, &peername) <= 0)
	error("error receiving message: %s\n", strerror(errno));

    /* who is it from? */


    hp = gethostbyaddr((char *)&peername.sin_addr, 
		       sizeof(peername.sin_addr), AF_INET);
    if(hp == NULL) error("gethostbyaddr: couldn't resolve addr %08X", 
			 peername.sin_addr);
    strcpy(remotehost, hp->h_name);

    debug_write(inmsg->cur);
    debug_write("-----\n");

    gethostname(myhost, 256);
    sprintf(line, "hostname %s\n", myhost);
    dgram_cat(outmsg, line);

    /* parse top of message */

#ifdef notyet
    if(sscanf(inmsg->cur, "dumpuser %s\n", remoteuser) != 1)
	goto baderr;
#endif
    dgram_eatline(inmsg);

    if(sscanf(inmsg->cur, "replyport %d\n", &replyport) != 1) goto baderr;
    dgram_eatline(inmsg);

    if(sscanf(inmsg->cur, "program %s\n", program) != 1) goto err;
    if(strcmp(program, "DUMP")) goto err;
    dgram_eatline(inmsg);

    if(sscanf(inmsg->cur, "handle %s\n", handle) != 1) goto err;
    sprintf(line, "handle %s\n", handle);
    dgram_cat(outmsg, line);
    dgram_eatline(inmsg);

    sprintf(line, "speed %s\n", SPEEDS);
    dgram_cat(outmsg, line);

    /* who am I now? */

    myuid = getuid();
    if((pwptr = getpwuid(myuid)) == NULL)
	error("error [getpwuid(%d) fails]", myuid);
    strcpy(localuser, pwptr->pw_name);

#ifdef notyet

    /* check remote access */

    sprintf(line, "remote (host %s user %s port %d) local (user %s)\n", 
	    remotehost, remoteuser, replyport, localuser);
    debug_write(line);
    if(replyport >= IPPORT_RESERVED)
	debug_write("(remote port not secure)\n");

    chdir(pwptr->pw_dir);
    close(2);
    userok = ruserok(remotehost, myuid == 0, remoteuser, localuser);
    dup2(1,2);
    chdir("/tmp");
    if(userok == -1) debug_write("(ruserok failed)\n");
    else debug_write("(ruserok passes)\n");
    debug_write("-----\n");

#endif

    /* handle all service requests */

    while(*inmsg->cur) {
	if(sscanf(inmsg->cur, "%s %s\n", disk, levels) != 2) goto err;
	dgram_eatline(inmsg);

	dgram_cat(outmsg, disk);
	for(lp = levels; *lp; lp++) {
	    size = getsize(disk, *lp);
	    sprintf(line, " %d", size);
	    dgram_cat(outmsg, line);
	}
	dgram_cat(outmsg, "\n");
    }

    debug_write(outmsg->data);
    debug_write("=====\n");

    dgram_send(remotehost, replyport, outmsg);
    exit(0);

 err:
    dgram_cat(outmsg, "ERROR in request packet\n");
    dgram_send(remotehost, replyport, outmsg);
 baderr:
    exit(1);
}


static int getsize(disk, level)
char *disk;
int level;
{
    int pipefd[2], nullfd, size, dumppid;
    FILE *dumpout;
    char dumpkeys[10], device[80];

    nullfd = open("/dev/null", O_RDWR);

    pipe(pipefd);

    switch(dumppid = fork()) {
    case -1: return -1;
    default: break; 
    case 0:	/* child process */
	if(SETPGRP == -1) {
	    char msg[100];
	    sprintf(msg,"setpgrp failed: %s\n",strerror(errno));
	    debug_write(msg);
	}
	dup2(nullfd, 0);
	dup2(nullfd, 1);
	dup2(pipefd[1], 2);
	close(pipefd[0]);

#ifndef AIX_BACKUP
	sprintf(dumpkeys, "%csf", level);
	sprintf(device, "%s/%s", DEVDIR, disk);
	execl(DUMP, "dump", dumpkeys, "100000", "-", device, 0);
#else
	sprintf(dumpkeys, "-%cf", level);
	sprintf(device, "%s/%s", DEVDIR, disk);
	execl(DUMP, "backup", dumpkeys, "-", device, 0);
#endif
	exit(1);
    }
    close(pipefd[1]);
    dumpout = fdopen(pipefd[0],"r");

    size = -1;
    while(fgets(line,MAXLINE,dumpout) != NULL) {
	debug_write(line);
	size = handle_dumpline(line);
	if(size > -1) break;
    }

    debug_write(".....\n");
    if(size == -1)
	debug_write("(no size line match in above dump output)\n.....\n");

    kill(-dumppid, SIGTERM);

    close(nullfd);
    fclose(dumpout);

    return size;
}


static int first_num(str)
char *str;
/*
 * Returns the value of the first integer in a string.
 */
{
    char tmp[16], *tp;

    tp = tmp;
    while(*str && !isdigit(*str)) str++;
    while(*str && isdigit(*str)) *tp++ = *str++;
    *tp = '\0';

    return atoi(tmp);
}


static int handle_dumpline(str)
char *str;
/*
 * Checks the dump output line against the error and size regex tables.
 */
{
    regex_t *rp;
    char *errs, *re_comp();
    
    /* check for size match */
    for(rp = re_size; rp->regex != NULL; rp++) {
	errs = re_comp(rp->regex);
	/* check for error return: errs != NULL */
	if(re_exec(str) == 1) 
	    return (first_num(str)*rp->scale+1023)/1024;
    }
    return -1;
}
